
/* 
Compile with:   
nvcc -Xcompiler '-fPIC' -arch sm_20 --ptxas-options=-v --shared -o calc-strat.so calc-strat.cu -lcudart -lcuda -L/usr/lib64/R/lib -lR

Add to .emacs
(add-to-list 'auto-mode-alist '("\\.cu$" . c++-mode))
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define CUDA_CALL(x) do { if((x) != cudaSuccess) { \
    printf("Error at %s:%d\n",__FILE__,__LINE__);\
    exit(EXIT_FAILURE);}} while(0)

void checkCUDAError(const char *msg, const int line)
{
    cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) 
    {
      printf("Cuda error: %s (line %d): %s.\n", msg, line, cudaGetErrorString( err) );
        exit(EXIT_FAILURE);
    }                         
}


#define MatrID       0
#define MatrRound    1
#define MatrHold     2
#define MatrPass     3
#define MatrMaxHold  4
#define MatrMaxPass  5
#define MatrLow      6
#define MatrUp       7
#define MatrStep     8
#define MatrRate     9
#define MatrNormHold 10
#define MatrNormPass 11   
#define MatrDim      12


#define ParLambda    0
#define ParRho       1
#define ParSigma     2
#define ParAlpha     3
#define ParBeta      4
#define AnzObs       5
#define AnzSimul     6
#define LSubsPars    7
#define Scheme       8
#define ParsLength   9

#define MaxStrats   1000



extern "C" void calcExpPay (double *data, double *pars, double *subsPars, double *choiceProbs);


//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void printMatrTransp (float *bidK, int dim1, int dim2, const char *s)
{
  int size1 = dim1*dim2*sizeof(float);

  float *bidF = (float *) malloc (size1);
  cudaError_t err = cudaMemcpy(bidF, bidK, size1, cudaMemcpyDeviceToHost);
  printf("CUDA ERROR (%d) {%s}\n", err, cudaGetErrorString(err));
  if (err) throw err;

  printf ("%s\n", s);
  int lim2 = dim2>10 ? 10 : dim2;
  for (int i1=0; i1<dim1; i1++) {
    for (int i2=0; i2<dim2; i2++) 
      if (i1<=9 && i2<=lim2) printf ("%4.4f  ", bidF[i2*dim1+i1]);
    if (i1<=10) printf ("\n");
  }
  printf ("\n");
  
  free (bidF);
}

void printMatrTranspDouble (double *bidK, int dim1, int dim2, const char *s)
{
  int size1 = dim1*dim2*sizeof(double);

  double *bidF = (double *) malloc (size1);
  cudaError_t err = cudaMemcpy(bidF, bidK, size1, cudaMemcpyDeviceToHost);
  printf("CUDA ERROR (%d) {%s}\n", err, cudaGetErrorString(err));
  if (err) throw err;

  printf ("%s\n", s);
  int lim2 = dim2>10 ? 10 : dim2;
  for (int i1=0; i1<dim1; i1++) {
    for (int i2=0; i2<dim2; i2++) 
      if (i1<=9 && i2<=lim2) printf ("%4.4f  ", bidF[i2*dim1+i1]);
    if (i1<=10) printf ("\n");
  }
  printf ("\n");
  
  free (bidF);
}

void printMatrDouble (double *bidK, int dim1, int dim2, const char *s)
{
  int size1 = dim1*dim2*sizeof(double);

  double *bidF = (double *) malloc (size1);
  cudaError_t err = cudaMemcpy(bidF, bidK, size1, cudaMemcpyDeviceToHost);
  printf("CUDA ERROR (%d) {%s}\n", err, cudaGetErrorString(err));
  if (err) throw err;

  printf ("%s\n", s);
  for (int i1=0; i1<dim1; i1++) {
    for (int i2=0; i2<dim2; i2++)
      if (i1<=9 && i2<=9) printf ("%4.4f  ", bidF[i1*dim2+i2]);
    if (i1<=10) printf ("\n");
  }
  printf ("\n");
  
  free (bidF);
}


//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////



__global__ void calcUtils (double *dataK, double *choiceProbsK, float *utilsK, float *parsK)
{

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx > parsK[AnzObs]-0.9f) return;


  double *d = dataK + MatrDim * idx ;
  float *u = utilsK + MaxStrats * idx ;
  
  int numStrats = (int) rint(d[MatrUp]);
  float alpha = parsK[ParAlpha] ;
  float beta = parsK[ParBeta] ;

  //  if (idx<3) printf ("idx=%d, id=%3.0f, round=%3.0f, numStrats=%d, alpha=%3.3f, beta=%3.3f\n", idx, d[MatrID], d[MatrRound], numStrats, alpha, beta) ;

  float maxU = -1e5f;
  for (int s=0; s<=numStrats; s++) {
    float me = (float) numStrats - s + 1.0f;
    float opp = (float) s * d[MatrRate] + 1.0f ;

    if (fabsf(beta)<1e-3f) me = powf(me, 1.0f-alpha) * powf (opp, alpha);
    else {
      me = (1.0f-alpha) * powf(me, beta) + alpha * powf(opp, beta);

      if (me < 0.0f) me = -powf(-me, 1.0f/beta);
      else me = powf(me, 1.0f/beta);
    }

    if (me>maxU) maxU = me;

    u[s] = me ;
  }

  for (int s=0; s<=numStrats; s++) u[s] -= maxU ;

}



__global__ void calcLogitProbs (double *dataK, double *choiceProbsK, float *utilsK, float *parsK)
{

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx > parsK[AnzObs]-0.9f) return;

  double *d = dataK + MatrDim * idx ;
  float *u = utilsK + MaxStrats * idx ;
  
  int numStrats = (int) rint(d[MatrUp]);
  float sum = 0.0f;
  float lambda = parsK[ParLambda] ;

  for (int s=0; s<=numStrats; s++) sum += expf (lambda * u[s]);

  int strat = (int) rint(d[MatrNormPass]);
  choiceProbsK[idx] = (double) expf (lambda * u[strat]) / sum ;
}



__device__ void calcS1GStratsNR (double *d, float *stratK, float *u, float *parsK, int numSteps, int currStrat, float currProb, int numStrats)
{

  int maxStrats = (int) rint(d[MatrUp]);

  int n1 = currStrat-max(currStrat-numStrats,0) ;
  int n2 = min(currStrat+numStrats,maxStrats)-currStrat ;

  int step=1;
  if (n1 / 100 + n2 / 100 >= numSteps-1) step=100;
  else if (n1 / 50 + n2 / 50 >= numSteps-1) step=50;
  else if (n1 / 10 + n2 / 10 >= numSteps-1) step=10;
  else if (n1 / 5 + n2 / 5 >= numSteps-1) step=5;

  n1 /= step;
  n2 /= step;

  float lambda = parsK[ParLambda] ;
  float rho = parsK[ParRho] ;

  float maxU = -1e6f;
  for (int s=currStrat-n1*step; s<=currStrat+n2*step; s+=step)
    if (u[s]>maxU) maxU = u[s] ;

  float sum = 0.0f;
  for (int s=currStrat-n1*step; s<=currStrat+n2*step; s+=step)
    sum += expf (lambda * (u[s]-maxU));

  for (int s=currStrat-n1*step; s<=currStrat+n2*step; s+=step) {
    float p = currProb * expf (lambda * (u[s]-maxU)) / sum ;
    stratK[s] += rho * p ;
    if (step>1) calcS1GStratsNR (d, stratK, u, parsK, numSteps, s, (1.0f-rho) * p, (int) step/2) ;
    else stratK[s] += (1.0f-rho) * p ;
  }

}


__global__ void calcS1GStrats (double *dataK, float *stratK, float *utilsK, float *parsK, int numSteps)
{

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx > parsK[AnzObs]-0.9f) return;

  double *d = dataK + MatrDim * idx ;
  float *u = utilsK + MaxStrats * idx ;
  float *s = stratK + MaxStrats * idx ;
  int maxStrats = (int) rint(d[MatrUp]);

  calcS1GStratsNR (d, s, u, parsK, numSteps, 0, 1.0f, maxStrats) ;
}


__global__ void weightS1GProbs (double *choiceProbsK, float *stratK1, float *stratK2, double *dataK, float *parsK, float weight)
{

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx > parsK[AnzObs]-0.9f) return;

  double *d = dataK + MatrDim * idx ;
  int strat = (int) rint(d[MatrNormPass]);

  float *s1 = stratK1 + MaxStrats * idx ;
  float *s2 = stratK2 + MaxStrats * idx ;

  int numStrats = (int) rint(d[MatrUp]);
  float sum = 0.0f;
  for (int s=0; s<=numStrats; s++) sum += weight * s1[s] + (1.0f-weight) * s2[s] ;

  choiceProbsK[idx] = (double) (weight * s1[strat] + (1.0f-weight) * s2[strat]) / sum ;
}


//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////


extern "C" void calcExpPay (double *data, double *pars, double *subsPars, double *choiceProbs) {

  if (GPU_Num > 0) {
    int devArr[2];
    devArr[0]=GPU_Num;
    devArr[1]=1-GPU_Num;
    cudaSetValidDevices(devArr, (int) 1);
    checkCUDAError ("Set devices",(__LINE__)-1);
  }

  int anzObs = (int) rint(pars[AnzObs]);
  int anzSimul = (int) rint(pars[AnzSimul]);
  int lSubsPars = (int) rint(pars[LSubsPars]);
  int scheme = (int) rint (pars[Scheme]) ;

  if (LOG_FL >= 1) printf ("Enter calcExpPay (AnzObs=%d, AnzSimul=%d, LSubsPars=%d, Scheme=%d)\n", anzObs, anzSimul, lSubsPars, scheme) ;

  int sizePars = ParsLength*sizeof(float);
  int sizeData = anzObs*MatrDim*sizeof(double);
  int sizeProbs = anzSimul*anzObs*sizeof(double);
  int sizeStrats = anzObs*MaxStrats*sizeof(float);
  int sizeUtils = MaxStrats*anzObs*sizeof(float);

  double *dataK, *choiceProbsK, *choiceProbsK2;
  float *parsK, *utilsK, *stratsK1, *stratsK2;
  float parsF[ParsLength];

  if (LOG_FL >= 1) printf ("Copy pars[]\n") ;
  CUDA_CALL(cudaMalloc ((void **) &parsK, sizePars));
  for (int i=0; i<ParsLength; i++) parsF[i] = (float) pars[i];
  CUDA_CALL(cudaMemcpy (parsK, parsF, sizePars, cudaMemcpyHostToDevice));

  if (LOG_FL >= 1) printf ("Alloc choiceProbsK[]\n") ;
  CUDA_CALL(cudaMalloc ((void **) &choiceProbsK, sizeProbs));
  CUDA_CALL(cudaMemset (choiceProbsK, 0, sizeProbs));

  if (LOG_FL >= 1) printf ("Alloc stratsK[]\n") ;
  CUDA_CALL(cudaMalloc ((void **) &stratsK1, sizeStrats));
  CUDA_CALL(cudaMalloc ((void **) &stratsK2, sizeStrats));

  if (LOG_FL >= 1) printf ("Alloc utilsK[]\n") ;
  CUDA_CALL(cudaMalloc ((void **) &utilsK, sizeUtils));
  CUDA_CALL(cudaMemset (utilsK, 0, sizeUtils));

  if (LOG_FL >= 1) printf ("Alloc dataK[]\n") ;
  CUDA_CALL(cudaMalloc ((void **) &dataK, sizeData));
  CUDA_CALL(cudaMemcpy (dataK, data, sizeData, cudaMemcpyHostToDevice));
  if (LOG_FL >= 2) printMatrDouble (dataK, (int) anzObs, (int) MatrDim, "dataK");

  int numBlocks = (int) anzObs / 512 + 1;
  choiceProbsK2 = choiceProbsK;

  for (int s=0; s<anzSimul; s++) {
    if (LOG_FL >= 1) printf ("Simul-Run %d\n", s);

    if (lSubsPars==2) {
      parsF[ParAlpha] = (float) subsPars[2*s];
      parsF[ParBeta] = (float) subsPars[2*s+1];
    } else {
      parsF[ParLambda] = (float) subsPars[3*s];
      parsF[ParAlpha] = (float) subsPars[3*s+1];
      parsF[ParBeta] = (float) subsPars[3*s+2];
    }
    CUDA_CALL(cudaMemcpy (parsK, parsF, sizePars, cudaMemcpyHostToDevice));

    if (LOG_FL >= 1) printf ("starting calcUtils(%d, 512)\n", numBlocks) ;
    calcUtils<<<numBlocks,512>>>(dataK, choiceProbsK2, utilsK, parsK);
    checkCUDAError ("calcUtils",(__LINE__)-1);
    if (LOG_FL >= 2) printMatrTransp (utilsK, (int) MaxStrats, (int) parsF[AnzObs], "utils");

    if (scheme==1) { // multinomial logit

      if (LOG_FL >= 1) printf ("starting calcLogitProbs()\n") ;
      calcLogitProbs<<<numBlocks,512>>>(dataK, choiceProbsK2, utilsK, parsK);
      checkCUDAError ("calcLogitProbs",(__LINE__)-1);
      if (LOG_FL >= 2) printMatrTranspDouble (choiceProbsK2, (int) 1, (int) parsF[AnzObs], "choiceProbs");
    } 
    else if (scheme==2) {  // general Grid Logit

      int numSteps1 = (int) floor(1.0/pars[ParSigma]);
      int numSteps2 = (int) ceil(1.0/pars[ParSigma]);
      float mixWeight = (float) ceil(1.0/pars[ParSigma]) - 1.0/pars[ParSigma] ;

      CUDA_CALL(cudaMemset (stratsK1, 0, sizeStrats));
      CUDA_CALL(cudaMemset (stratsK2, 0, sizeStrats));

      if (LOG_FL >= 1) printf ("starting calcS1GStrats(%d)\n", numSteps1) ;
      calcS1GStrats<<<numBlocks,512>>>(dataK, stratsK1, utilsK, parsK, numSteps1);
      checkCUDAError ("calcS1GStrats",(__LINE__)-1);
      if (LOG_FL >= 2) printMatrTransp (stratsK1, (int) MaxStrats, (int) parsF[AnzObs], "stratsK1");

      if (LOG_FL >= 1) printf ("starting calcS1GStrats(%d)\n", numSteps2) ;
      calcS1GStrats<<<numBlocks,512>>>(dataK, stratsK2, utilsK, parsK, numSteps2);
      checkCUDAError ("calcS1GStrats",(__LINE__)-1);
      if (LOG_FL >= 2) printMatrTransp (stratsK2, (int) MaxStrats, (int) parsF[AnzObs], "stratsK2");
      
      if (LOG_FL >= 1) printf ("starting weightS1GProbs(%3.3f)\n", mixWeight) ;
      weightS1GProbs<<<numBlocks,512>>>(choiceProbsK2, stratsK1, stratsK2, dataK, parsK, mixWeight);
      checkCUDAError ("weightS1GProbs",(__LINE__)-1);
      if (LOG_FL >= 2) printMatrTranspDouble (choiceProbsK2, (int) 1, (int) parsF[AnzObs], "choiceProbs");
      
    }

    choiceProbsK2 += anzObs ;
  }

  CUDA_CALL(cudaMemcpy(choiceProbs, choiceProbsK, sizeProbs, cudaMemcpyDeviceToHost));

  CUDA_CALL(cudaFree (parsK));
  CUDA_CALL(cudaFree (choiceProbsK));
  CUDA_CALL(cudaFree (stratsK1));
  CUDA_CALL(cudaFree (stratsK2));
  CUDA_CALL(cudaFree (dataK));
  CUDA_CALL(cudaFree (utilsK));

//  cudaThreadExit();
}
