
/* 
Compile with:   
nvcc -Xcompiler '-fPIC' -arch sm_20 --ptxas-options=-v --shared -o calc-strat.so calc-strat.cu -lcudart -lcuda -L/usr/lib64/R/lib -lR

Add to .emacs
(add-to-list 'auto-mode-alist '("\\.cu$" . c++-mode))
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define CUDA_CALL(x) do { if((x) != cudaSuccess) { \
    printf("Error at %s:%d\n",__FILE__,__LINE__);\
    exit(EXIT_FAILURE);}} while(0)

void checkCUDAError(const char *msg, const int line)
{
    cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) 
    {
      printf("Cuda error: %s (line %d): %s.\n", msg, line, cudaGetErrorString( err) );
        exit(EXIT_FAILURE);
    }                         
}


#define MatrID       0
#define MatrRound    1
#define MatrHold     2
#define MatrPass     3
#define MatrMaxHold  4
#define MatrMaxPass  5
#define MatrLow      6
#define MatrUp       7
#define MatrStep     8
#define MatrRate     9
#define MatrNormHold 10
#define MatrNormPass 11   
#define MatrDim      12


#define ParLambda    0
#define ParRho       1
#define ParSigma     2
#define ParAlpha     3
#define ParBeta      4
#define AnzObs       5
#define LSubsPars    6
#define DimSimul     7
#define Scheme       8
#define Moderate     9
#define ParsLength   10

#define MaxStrats   1000



extern "C" void calcExpPay (double *data, double *pars, double *subsPars, double *choiceProbs);


//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void printMatrTransp (float *bidK, int dim1, int dim2, const char *s)
{
  int size1 = dim1*dim2*sizeof(float);

  float *bidF = (float *) malloc (size1);
  cudaError_t err = cudaMemcpy(bidF, bidK, size1, cudaMemcpyDeviceToHost);
  printf("CUDA ERROR (%d) {%s}\n", err, cudaGetErrorString(err));
  if (err) throw err;

  printf ("%s\n", s);
  int lim2 = dim2>10 ? 10 : dim2;
  for (int i1=0; i1<dim1; i1++) {
    for (int i2=0; i2<dim2; i2++) 
      if (i1<=9 && i2<=lim2) printf ("%4.4f  ", bidF[i2*dim1+i1]);
    if (i1<=10) printf ("\n");
  }
  printf ("\n");
  
  free (bidF);
}

void printMatrTranspDouble (double *bidK, int dim1, int dim2, const char *s)
{
  int size1 = dim1*dim2*sizeof(double);

  double *bidF = (double *) malloc (size1);
  cudaError_t err = cudaMemcpy(bidF, bidK, size1, cudaMemcpyDeviceToHost);
  printf("CUDA ERROR (%d) {%s}\n", err, cudaGetErrorString(err));
  if (err) throw err;

  printf ("%s\n", s);
  int lim2 = dim2>10 ? 10 : dim2;
  for (int i1=0; i1<dim1; i1++) {
    for (int i2=0; i2<dim2; i2++) 
      if (i1<=9 && i2<=lim2) printf ("%4.4f  ", bidF[i2*dim1+i1]);
    if (i1<=10) printf ("\n");
  }
  printf ("\n");
  
  free (bidF);
}

void printMatrDouble (double *bidK, int dim1, int dim2, const char *s)
{
  int size1 = dim1*dim2*sizeof(double);

  double *bidF = (double *) malloc (size1);
  cudaError_t err = cudaMemcpy(bidF, bidK, size1, cudaMemcpyDeviceToHost);
  printf("CUDA ERROR (%d) {%s}\n", err, cudaGetErrorString(err));
  if (err) throw err;

  printf ("%s\n", s);
  for (int i1=0; i1<dim1; i1++) {
    for (int i2=0; i2<dim2; i2++)
      if (i1<=9 && i2<=9) printf ("%4.4f  ", bidF[i1*dim2+i2]);
    if (i1<=10) printf ("\n");
  }
  printf ("\n");
  
  free (bidF);
}


//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////



__global__ void calcUtils (double *dataK, double *choiceProbsK, float *utilsK, float *parsK)
{

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx > parsK[AnzObs]-0.9f) return;


  double *d = dataK + MatrDim * idx ;
  float *u = utilsK + MaxStrats * idx ;
  
  int numStrats = (int) rint(d[MatrUp]);
  float alpha = parsK[ParAlpha] ;
  float beta = parsK[ParBeta] ;
  int moderate = (int) rint (parsK[Moderate]) ;


  //  if (idx<3) printf ("idx=%d, id=%3.0f, round=%3.0f, numStrats=%d, alpha=%3.3f, beta=%3.3f\n", idx, d[MatrID], d[MatrRound], numStrats, alpha, beta) ;

  float maxU = -1e5f;
  float minU = 1e5f;
  for (int s=0; s<=numStrats; s++) {
    float me = (float) numStrats - s;
    float opp = (float) s * d[MatrRate];

    if (moderate==2) {
      me /= 30.0f;
      opp /= 30.0f;
    }

    me += 1.0f;
    opp += 1.0f;

    if (fabsf(beta)<1e-3f) me = powf(me, 1.0f-alpha) * powf (opp, alpha);
    else {
      me = (1.0f-alpha) * powf(me, beta) + alpha * powf(opp, beta);

      if (me < 0.0f) me = -powf(-me, 1.0f/beta);
      else me = powf(me, 1.0f/beta);
    }

    if (me>maxU) maxU = me;
    if (me<minU) minU = me;

    u[s] = me ;
  }

  float dU = maxU-minU;
  if (maxU-minU < 1e-3f) dU = 1e-3f;
  int scheme = (int) rint (parsK[Scheme]) ;

  if (scheme < 3 || scheme % 2==1) {
    for (int s=0; s<=numStrats; s++) u[s] -= maxU ;
  } else {
    for (int s=0; s<=numStrats; s++) { u[s] /= dU ; u[s] -= maxU/dU ; }
  }

}



__global__ void calcLogitProbs (double *dataK, double *choiceProbsK, float *utilsK, float *parsK)
{

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx > parsK[AnzObs]-0.9f) return;

  double *d = dataK + MatrDim * idx ;
  float *u = utilsK + MaxStrats * idx ;
  
  int numStrats = (int) rint(d[MatrUp]);
  float sum = 0.0f;
  float lambda = parsK[ParLambda] ;

  for (int s=0; s<=numStrats; s++) sum += expf (lambda * u[s]);

  int strat = (int) rint(d[MatrNormPass]);
  choiceProbsK[idx] = (double) expf (lambda * u[strat]) / sum ;
}

// For simplicity assume sigma=1 here

__global__ void calcOleProbs (double *dataK, double *choiceProbsK, float *utilsK, float *parsK)
{

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx > parsK[AnzObs]-0.9f) return;

  double *d = dataK + MatrDim * idx ;
  float *u = utilsK + MaxStrats * idx ;
  
  int numStrats = (int) rint(d[MatrUp]);
  float lambda = parsK[ParLambda] ;
  float rho = parsK[ParRho] ;

  // w <- dnorm (-4:4)/ sum(dnorm (-4:4)) ; w - log(sum(exp(w)))
  float w[9] = { -2.318115f,-2.313817f,-2.264257f,-2.076277f,-1.919305f,-2.076277f,-2.264257f,-2.313817f,-2.318115f };

  float maxI = -1e15f;
  for (int r=0; r<=numStrats+8; r++) {

    float maxB = -1e15f;
    for (int j=max(0, r-8); j<=min(r,numStrats); j++) {
      float b = lambda*u[j] / rho + w[r-j];
      if (b > maxB) maxB = b;
    }

    float sumB = 0.0f;
    for (int j=max(0, r-8); j<=min(r,numStrats); j++) {
      float b = lambda*u[j] / rho + w[r-j];
      sumB += expf (b - maxB);
    }

    float Ir = logf(sumB) + maxB;
    if (Ir > maxI) maxI = Ir;
  }


  float sumQ = 0.0f;
  for (int r=0; r<=numStrats+8; r++) {
    float maxB = -1e15f;
    for (int j=max(0, r-8); j<=min(r,numStrats); j++) {
      float b = lambda*u[j] / rho + w[r-j];
      if (b > maxB) maxB = b;
    }

    float sumB = 0.0f;
    for (int j=max(0, r-8); j<=min(r,numStrats); j++) {
      float b = lambda*u[j] / rho + w[r-j];
      sumB += expf (b - maxB);
    }

    float Ir = logf(sumB) + maxB;
    sumQ += expf (rho*(Ir-maxI));
  }


  int strat = (int) rint(d[MatrNormPass]);
  float prob = 0.0f;
  for (int r=strat; r<=strat+8; r++) {
    float maxB = -1e15f;
    for (int j=max(0, r-8); j<=min(r,numStrats); j++) {
      float b = lambda*u[j] / rho + w[r-j];
      if (b > maxB) maxB = b;
    }

    float sumB = 0.0f;
    for (int j=max(0, r-8); j<=min(r,numStrats); j++) {
      float b = lambda*u[j] / rho + w[r-j];
      sumB += expf (b - maxB);
    }

    float Ir = logf(sumB) + maxB;
    float q = expf (lambda*u[strat] / rho + w[r-strat] - Ir);

    float Qr = expf (rho*(Ir-maxI)) / sumQ;
    prob += q * Qr;
  }

  choiceProbsK[idx] = (double) prob;
}


__global__ void calcPromProbs (double *dataK, double *choiceProbsK, float *utilsK, float *parsK)
{

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx > parsK[AnzObs]-0.9f) return;

  double *d = dataK + MatrDim * idx ;
  float *u = utilsK + MaxStrats * idx ;
  
  int numStrats = (int) rint(d[MatrUp]);
  float sum = 0.0f;
  float lambda = parsK[ParLambda] ;
  float rho = parsK[ParRho] ;

  int lev = 0;
  for (int s=0; s<=numStrats; s++) {
    if (s==0) lev = 2;
    else if (s%5000==0) lev = 5;
    else if (s%1000==0) lev = 4;
    else if (s%500==0) lev = 3;
    else if (s%100==0) lev = 2;
    else if (s%50==0) lev = 1;
    else if (s%10==0) lev = 0;
    else if (s%5==0) lev = -1;
    else lev = -2;

    sum += expf (lambda * u[s] + rho * lev);
  }

  int strat = (int) rint(d[MatrNormPass]);

  if (strat==0) lev = 2;
  else if (strat % 5000==0) lev = 5;
  else if (strat % 1000==0) lev = 4;
  else if (strat % 500==0) lev = 3;
  else if (strat % 100==0) lev = 2;
  else if (strat % 50==0) lev = 1;
  else if (strat % 10==0) lev = 0;
  else if (strat % 5==0) lev = -1;
  else lev = -2;

  choiceProbsK[idx] = (double) expf (lambda * u[strat] + rho * lev) / sum ;
}


__global__ void calcHetProbs (double *dataK, double *choiceProbsK, float *utilsK, float *parsK)
{

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx > parsK[AnzObs]-0.9f) return;

  double *d = dataK + MatrDim * idx ;
  float *u = utilsK + MaxStrats * idx ;
  
  int numStrats = (int) rint(d[MatrUp]);
  float sum = 0.0f;
  float lambda = parsK[ParLambda] ;
  float rho = parsK[ParRho] ;

  int maxLev = 0;
  if (numStrats > 5000) maxLev = 5;
  else if (numStrats > 1000) maxLev = 4;
  else if (numStrats > 500) maxLev = 3;
  else if (numStrats > 100) maxLev = 2;
  else if (numStrats > 50) maxLev = 1;
  else if (numStrats > 10) maxLev = 0;
  else if (numStrats > 5) maxLev = -1;
  else maxLev = -2;

  int lev = 0;
  for (int s=0; s<=numStrats; s++) {
    if (s==0) lev = maxLev;
    else if (s%5000==0) lev = 5;
    else if (s%1000==0) lev = 4;
    else if (s%500==0) lev = 3;
    else if (s%100==0) lev = 2;
    else if (s%50==0) lev = 1;
    else if (s%10==0) lev = 0;
    else if (s%5==0) lev = -1;
    else lev = -2;

    sum += expf (lambda * u[s] + rho * lev);
  }

  int strat = (int) rint(d[MatrNormPass]);

  if (strat==0) lev = maxLev;
  else if (strat % 5000==0) lev = 5;
  else if (strat % 1000==0) lev = 4;
  else if (strat % 500==0) lev = 3;
  else if (strat % 100==0) lev = 2;
  else if (strat % 50==0) lev = 1;
  else if (strat % 10==0) lev = 0;
  else if (strat % 5==0) lev = -1;
  else lev = -2;

  choiceProbsK[idx] = (double) expf (lambda * u[strat] + rho * lev) / sum ;
}

__global__ void calcPalmProbs (double *dataK, double *choiceProbsK, float *utilsK, float *parsK)
{

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx > parsK[AnzObs]-0.9f) return;

  double *d = dataK + MatrDim * idx ;
  float *u = utilsK + MaxStrats * idx ;
  
  int numStrats = (int) rint(d[MatrUp]);
  float sum = 0.0f;
  float lambda = parsK[ParLambda] ;
  float rho = parsK[ParRho] ;

  for (int s=0; s<=numStrats; s++) sum += expf (lambda * u[s]);

  float prevLevProb = 1.0f;
  float sumProb = 0.0f;
  int strat = (int) rint(d[MatrNormPass]);
  float stratProb = 0.0f;

  // loop over levels
  //    loop over strats
  //       if str on lev, calc prob, add to thisLevProb and to sumProb
  //       if str==strat, save prob in stratProb
  //

  int prevLevStep = 100000;

  for (int lev=5; lev>=-2; lev--) {
    int thisLevStep = 1;
    if (lev==5) thisLevStep = 5000;
    if (lev==4) thisLevStep = 1000;
    if (lev==3) thisLevStep = 500;
    if (lev==2) thisLevStep = 100;
    if (lev==1) thisLevStep = 50;
    if (lev==0) thisLevStep = 10;
    if (lev==-1) thisLevStep = 5;

    float thisLevProb = 0.0f;
    int loopEnd = rintf(numStrats/thisLevStep);
    int loopStart = 1;
    if (lev==2) loopStart = 0;
    
    if (numStrats > loopStart*thisLevStep)
      for (int s=loopStart; s<=loopEnd; s++) {
	int str = s*thisLevStep;
	if (s>0 && str % prevLevStep == 0) continue;

	float strProb = expf (lambda * u[str]) / sum;
	thisLevProb += strProb;
	sumProb += strProb * prevLevProb;
	if (str==strat) stratProb = strProb * prevLevProb;
      }

    prevLevProb *= fmaxf (0.0f, 1.0f - rho * thisLevProb);
    prevLevStep = thisLevStep;
  }

  choiceProbsK[idx] = (double) (stratProb / sumProb);
}



__global__ void calcPalm2Probs (double *dataK, double *choiceProbsK, float *utilsK, float *parsK)
{

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx > parsK[AnzObs]-0.9f) return;

  double *d = dataK + MatrDim * idx ;
  float *u = utilsK + MaxStrats * idx ;
  
  int numStrats = (int) rint(d[MatrUp]);
  float sum = 0.0f;
  float lambda = parsK[ParLambda] ;
  float rho = parsK[ParRho] ;

  for (int s=0; s<=numStrats; s++) sum += expf (lambda * u[s]);

  float prevLevProb = 1.0f;
  float sumProb = 0.0f;
  int strat = (int) rint(d[MatrNormPass]);
  float stratProb = 0.0f;

  // loop over levels
  //    loop over strats
  //       if str on lev, calc prob, add to thisLevProb and to sumProb
  //       if str==strat, save prob in stratProb
  //

  int prevLevStep = 100000;

  for (int lev=5; lev>=-2; lev--) {
    int thisLevStep = 1;
    if (lev==5) thisLevStep = 5000;
    if (lev==4) thisLevStep = 1000;
    if (lev==3) thisLevStep = 500;
    if (lev==2) thisLevStep = 100;
    if (lev==1) thisLevStep = 50;
    if (lev==0) thisLevStep = 10;
    if (lev==-1) thisLevStep = 5;

    float thisLevProb = 0.0f;
    int loopEnd = rintf(numStrats/thisLevStep);
    int loopStart = 1;
    if (lev==2) loopStart = 0;
    
    if (numStrats > loopStart*thisLevStep)
      for (int s=loopStart; s<=loopEnd; s++) {
	int str = s*thisLevStep;
	if (s>0 && (str % prevLevStep == 0)) continue;

	float strProb = expf (lambda * u[str]) / sum;
	thisLevProb += strProb;
	sumProb += strProb * prevLevProb;
	if (str==strat) stratProb = strProb * prevLevProb;
      }

    prevLevProb *= powf (1.0f - thisLevProb, rho);
    prevLevStep = thisLevStep;

    //    if (strat==0 && idx<1 && numStrats<100 && idx<100 && lev==-2) 
    // printf ("strat %d, prob %3.3f :\tlev %d, \tthisLevProb %3.3f, \tprevLevProb %3.3f  \t(idx %d, rho %3.3f, numStrats %d)\n", strat, stratProb / sumProb, lev, thisLevProb, prevLevProb, idx, rho, numStrats);
  }

  choiceProbsK[idx] = (double) (stratProb / sumProb);
}


__global__ void calcPalm3Probs (double *dataK, double *choiceProbsK, float *utilsK, float *parsK)
{

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx > parsK[AnzObs]-0.9f) return;

  double *d = dataK + MatrDim * idx ;
  float *u = utilsK + MaxStrats * idx ;
  
  int numStrats = (int) rint(d[MatrUp]);
  float sum = 0.0f;
  float lambda = parsK[ParLambda] ;
  float rho = fminf (1.0f, parsK[ParRho]) ;

  for (int s=0; s<=numStrats; s++) sum += expf (lambda * u[s]);

  float prevLevProb = 1.0f;
  float sumProb = 0.0f;
  int strat = (int) rint(d[MatrNormPass]);
  float stratProb = 0.0f;

  // loop over levels
  //    loop over strats
  //       if str on lev, calc prob, add to thisLevProb and to sumProb
  //       if str==strat, save prob in stratProb
  //

  int prevLevStep = 100000;

  for (int lev=5; lev>=-2; lev--) {
    int thisLevStep = 1;
    if (lev==5) thisLevStep = 5000;
    if (lev==4) thisLevStep = 1000;
    if (lev==3) thisLevStep = 500;
    if (lev==2) thisLevStep = 100;
    if (lev==1) thisLevStep = 50;
    if (lev==0) thisLevStep = 10;
    if (lev==-1) thisLevStep = 5;

    float thisLevProb = 0.0f;
    int loopEnd = rintf(numStrats/thisLevStep);
    int loopStart = 1;
    if (lev==2) loopStart = 0;  //  || (lev==1 && numStrats<100)
    
    if (numStrats > loopStart*thisLevStep)
      for (int s=loopStart; s<=loopEnd; s++) {
	int str = s*thisLevStep;
	if (s>0 && str % prevLevStep == 0) continue;

	//	if (idx==1) printf ("lev %d: %d  (%d)\n", lev, str, numStrats);

	float strProb = expf (lambda * u[str]) / sum;
	thisLevProb += strProb;
	sumProb += strProb * prevLevProb;
	if (str==strat) stratProb = strProb * prevLevProb;
      }

    prevLevProb *= 1.0f - rho * thisLevProb;
    prevLevStep = thisLevStep;
  }

  choiceProbsK[idx] = (double) (stratProb / sumProb);
}



__device__ void calcS1GStratsNR (double *d, float *stratK, float *u, float *parsK, int numSteps, int currStrat, float currProb, int numStrats)
{

  int maxStrats = (int) rint(d[MatrUp]);

  int n1 = currStrat-max(currStrat-numStrats,0) ;
  int n2 = min(currStrat+numStrats,maxStrats)-currStrat ;

  int step=1;
  if (n1 / 100 + n2 / 100 >= numSteps-1) step=100;
  else if (n1 / 50 + n2 / 50 >= numSteps-1) step=50;
  else if (n1 / 10 + n2 / 10 >= numSteps-1) step=10;
  else if (n1 / 5 + n2 / 5 >= numSteps-1) step=5;

  n1 /= step;
  n2 /= step;

  float lambda = parsK[ParLambda] ;
  float rho = fminf (1.0f, parsK[ParRho]) ;

  float maxU = -1e6f;
  for (int s=currStrat-n1*step; s<=currStrat+n2*step; s+=step)
    if (u[s]>maxU) maxU = u[s] ;

  float sum = 0.0f;
  for (int s=currStrat-n1*step; s<=currStrat+n2*step; s+=step)
    sum += expf (lambda * (u[s]-maxU));

  for (int s=currStrat-n1*step; s<=currStrat+n2*step; s+=step) {
    float p = currProb * expf (lambda * (u[s]-maxU)) / sum ;
    stratK[s] += rho * p ;
    if (step>1) calcS1GStratsNR (d, stratK, u, parsK, numSteps, s, (1.0f-rho) * p, (int) step/2) ;
    else stratK[s] += (1.0f-rho) * p ;
  }

}


__global__ void calcS1GStrats (double *dataK, float *stratK, float *utilsK, float *parsK, int numSteps)
{

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx > parsK[AnzObs]-0.9f) return;

  double *d = dataK + MatrDim * idx ;
  float *u = utilsK + MaxStrats * idx ;
  float *s = stratK + MaxStrats * idx ;
  int maxStrats = (int) rint(d[MatrUp]);

  calcS1GStratsNR (d, s, u, parsK, numSteps, 0, 1.0f, maxStrats) ;
}


__global__ void weightS1GProbs (double *choiceProbsK, float *stratK1, float *stratK2, double *dataK, float *parsK, float weight)
{

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx > parsK[AnzObs]-0.9f) return;

  double *d = dataK + MatrDim * idx ;
  int strat = (int) rint(d[MatrNormPass]);

  float *s1 = stratK1 + MaxStrats * idx ;
  float *s2 = stratK2 + MaxStrats * idx ;

  int numStrats = (int) rint(d[MatrUp]);
  float sum = 0.0f;
  for (int s=0; s<=numStrats; s++) sum += weight * s1[s] + (1.0f-weight) * s2[s] ;

  choiceProbsK[idx] = (double) (weight * s1[strat] + (1.0f-weight) * s2[strat]) / sum ;
}


//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////


extern "C" void calcExpPay (double *data, double *pars, double *subsPars, double *choiceProbs) {

  if (GPU_Num > 0) {
    int devArr[2];
    devArr[0]=GPU_Num;
    devArr[1]=1-GPU_Num;
    cudaSetValidDevices(devArr, (int) 1);
    checkCUDAError ("Set devices",(__LINE__)-1);
  }

  if (LOG_FL >= 1) printf ("Enter calcExpPay\n") ;

  int anzObs = (int) rint(pars[AnzObs]);
  int lSubsPars = (int) rint(pars[LSubsPars]);
  int dimSimul = (int) rint(pars[DimSimul]);

  int sizePars = ParsLength*sizeof(float);
  int sizeData = anzObs*MatrDim*sizeof(double);
  int sizeProbs = lSubsPars*anzObs*sizeof(double);
  int sizeStrats = anzObs*MaxStrats*sizeof(float);
  int sizeUtils = MaxStrats*anzObs*sizeof(float);

  int scheme = (int) rint (pars[Scheme]) ;

  double *dataK, *choiceProbsK, *choiceProbsK2;
  float *parsK, *utilsK, *stratsK1, *stratsK2;
  float parsF[ParsLength];

  if (LOG_FL >= 1) printf ("Copy pars[]\n") ;
  CUDA_CALL(cudaMalloc ((void **) &parsK, sizePars));
  for (int i=0; i<ParsLength; i++) parsF[i] = (float) pars[i];
  CUDA_CALL(cudaMemcpy (parsK, parsF, sizePars, cudaMemcpyHostToDevice));

  if (LOG_FL >= 1) printf ("Alloc choiceProbsK[]\n") ;
  CUDA_CALL(cudaMalloc ((void **) &choiceProbsK, sizeProbs));
  CUDA_CALL(cudaMemset (choiceProbsK, 0, sizeProbs));

  if (LOG_FL >= 1) printf ("Alloc stratsK[]\n") ;
  CUDA_CALL(cudaMalloc ((void **) &stratsK1, sizeStrats));
  CUDA_CALL(cudaMalloc ((void **) &stratsK2, sizeStrats));

  if (LOG_FL >= 1) printf ("Alloc utilsK[]\n") ;
  CUDA_CALL(cudaMalloc ((void **) &utilsK, sizeUtils));
  CUDA_CALL(cudaMemset (utilsK, 0, sizeUtils));

  if (LOG_FL >= 1) printf ("Alloc dataK[]\n") ;
  CUDA_CALL(cudaMalloc ((void **) &dataK, sizeData));
  CUDA_CALL(cudaMemcpy (dataK, data, sizeData, cudaMemcpyHostToDevice));
  if (LOG_FL >= 2) printMatrDouble (dataK, (int) anzObs, (int) MatrDim, "dataK");

  int numBlocks = (int) anzObs / 512 + 1;
  choiceProbsK2 = choiceProbsK;

  for (int s=0; s<lSubsPars; s++) {
    if (LOG_FL >= 1) printf ("Simul-Run %d\n", s);

    if (dimSimul == 2) {
      parsF[ParAlpha] = (float) subsPars[2*s];
      parsF[ParBeta] = (float) subsPars[2*s+1];
    } else if (dimSimul == 3) {
      parsF[ParLambda] = (float) subsPars[3*s];
      parsF[ParAlpha] = (float) subsPars[3*s+1];
      parsF[ParBeta] = (float) subsPars[3*s+2];
    } else if (dimSimul == 4) {
      parsF[ParLambda] = (float) subsPars[4*s];
      parsF[ParRho] = (float) subsPars[4*s+1];
      parsF[ParAlpha] = (float) subsPars[4*s+2];
      parsF[ParBeta] = (float) subsPars[4*s+3];
    }

    CUDA_CALL(cudaMemcpy (parsK, parsF, sizePars, cudaMemcpyHostToDevice));

    if (LOG_FL >= 1) printf ("starting calcUtils(%d, 512)\n", numBlocks) ;
    calcUtils<<<numBlocks,512>>>(dataK, choiceProbsK2, utilsK, parsK);
    checkCUDAError ("calcUtils",(__LINE__)-1);
    if (LOG_FL >= 2) printMatrTransp (utilsK, (int) MaxStrats, (int) parsF[AnzObs], "utils");

    if (scheme==1 || scheme==3) { // multinomial logit  or moderate logit

      if (LOG_FL >= 1) printf ("starting calcLogitProbs()\n") ;
      calcLogitProbs<<<numBlocks,512>>>(dataK, choiceProbsK2, utilsK, parsK);
      checkCUDAError ("calcLogitProbs",(__LINE__)-1);
      if (LOG_FL >= 2) printMatrTranspDouble (choiceProbsK2, (int) 1, (int) parsF[AnzObs], "choiceProbs");
    } 
    else if (scheme==2 || scheme==4) {  // general Grid Logit  or moderate Grid Logit

      int numSteps1 = (int) floor(1.0/pars[ParSigma]);
      int numSteps2 = (int) ceil(1.0/pars[ParSigma]);
      float mixWeight = (float) ceil(1.0/pars[ParSigma]) - 1.0/pars[ParSigma] ;

      CUDA_CALL(cudaMemset (stratsK1, 0, sizeStrats));
      CUDA_CALL(cudaMemset (stratsK2, 0, sizeStrats));

      if (LOG_FL >= 1) printf ("starting calcS1GStrats(%d)\n", numSteps1) ;
      calcS1GStrats<<<numBlocks,512>>>(dataK, stratsK1, utilsK, parsK, numSteps1);
      checkCUDAError ("calcS1GStrats",(__LINE__)-1);
      if (LOG_FL >= 2) printMatrTransp (stratsK1, (int) MaxStrats, (int) parsF[AnzObs], "stratsK1");

      if (LOG_FL >= 1) printf ("starting calcS1GStrats(%d)\n", numSteps2) ;
      calcS1GStrats<<<numBlocks,512>>>(dataK, stratsK2, utilsK, parsK, numSteps2);
      checkCUDAError ("calcS1GStrats",(__LINE__)-1);
      if (LOG_FL >= 2) printMatrTransp (stratsK2, (int) MaxStrats, (int) parsF[AnzObs], "stratsK2");
      
      if (LOG_FL >= 1) printf ("starting weightS1GProbs(%3.3f)\n", mixWeight) ;
      weightS1GProbs<<<numBlocks,512>>>(choiceProbsK2, stratsK1, stratsK2, dataK, parsK, mixWeight);
      checkCUDAError ("weightS1GProbs",(__LINE__)-1);
      if (LOG_FL >= 2) printMatrTranspDouble (choiceProbsK2, (int) 1, (int) parsF[AnzObs], "choiceProbs");
      
    }
    else if (scheme==5 || scheme==6) { // prom logit or moderate prom logit

      if (LOG_FL >= 1) printf ("starting calcPromProbs()\n") ;
      calcPromProbs<<<numBlocks,512>>>(dataK, choiceProbsK2, utilsK, parsK);
      checkCUDAError ("calcPromProbs",(__LINE__)-1);
      if (LOG_FL >= 2) printMatrTranspDouble (choiceProbsK2, (int) 1, (int) parsF[AnzObs], "choiceProbs");

    }
    else if (scheme==7 || scheme==8) { // het logit or moderate het logit

      if (LOG_FL >= 1) printf ("starting calcHetProbs()\n") ;
      calcHetProbs<<<numBlocks,512>>>(dataK, choiceProbsK2, utilsK, parsK);
      checkCUDAError ("calcHetProbs",(__LINE__)-1);
      if (LOG_FL >= 2) printMatrTranspDouble (choiceProbsK2, (int) 1, (int) parsF[AnzObs], "choiceProbs");

    }
    else if (scheme==9 || scheme==10) { // palm2 logit or moderate palm2 logit

      if (LOG_FL >= 1) printf ("starting calcPalm2Probs()\n") ;
      calcPalm2Probs<<<numBlocks,512>>>(dataK, choiceProbsK2, utilsK, parsK);
      checkCUDAError ("calcPalm2Probs",(__LINE__)-1);
      if (LOG_FL >= 2) printMatrTranspDouble (choiceProbsK2, (int) 1, (int) parsF[AnzObs], "choiceProbs");

    }
    else if (scheme==11 || scheme==12) { // palm3 logit or moderate palm3 logit

      if (LOG_FL >= 1) printf ("starting calcPalm3Probs()\n") ;
      calcPalm3Probs<<<numBlocks,512>>>(dataK, choiceProbsK2, utilsK, parsK);
      checkCUDAError ("calcPalm3Probs",(__LINE__)-1);
      if (LOG_FL >= 2) printMatrTranspDouble (choiceProbsK2, (int) 1, (int) parsF[AnzObs], "choiceProbs");
      
    }
    else if (scheme==13 || scheme==14) { // ole logit or moderate ole logit

      if (LOG_FL >= 1) printf ("starting calcOleProbs()\n") ;
      calcOleProbs<<<numBlocks,512>>>(dataK, choiceProbsK2, utilsK, parsK);
      checkCUDAError ("calcOleProbs",(__LINE__)-1);
      if (LOG_FL >= 2) printMatrTranspDouble (choiceProbsK2, (int) 1, (int) parsF[AnzObs], "choiceProbs");

    }
    else if (scheme==15 || scheme==16) { // flexible palm logit or moderate palm logit

      if (LOG_FL >= 1) printf ("starting calcPalmProbs()\n") ;
      calcPalmProbs<<<numBlocks,512>>>(dataK, choiceProbsK2, utilsK, parsK);
      checkCUDAError ("calcPalmProbs",(__LINE__)-1);
      if (LOG_FL >= 2) printMatrTranspDouble (choiceProbsK2, (int) 1, (int) parsF[AnzObs], "choiceProbs");
      
    }


    choiceProbsK2 += anzObs ;
  }

  CUDA_CALL(cudaMemcpy(choiceProbs, choiceProbsK, sizeProbs, cudaMemcpyDeviceToHost));

  CUDA_CALL(cudaFree (parsK));
  CUDA_CALL(cudaFree (choiceProbsK));
  CUDA_CALL(cudaFree (stratsK1));
  CUDA_CALL(cudaFree (stratsK2));
  CUDA_CALL(cudaFree (dataK));
  CUDA_CALL(cudaFree (utilsK));

//  cudaThreadExit();
}
